#include "semaphores.h"
#include "device.h"
#include "utils.h"

#include <runtime/atomics.h>
#include <stdlib.h>

rt_create_vk_semaphore_pool_result rtCreateVkSemaphorePool(rt_vk_device *dev, uint32_t initial_size) {
    rt_vk_semaphore_pool sem_pool;
    rt_create_rwlock_result lock_res = rtCreateRWLock();
    if (!lock_res.ok) {
        return (rt_create_vk_semaphore_pool_result){.result = RT_UNKNOWN_ERROR};
    }
    sem_pool.resize_lock = lock_res.lock;
    sem_pool.dev = dev;
    sem_pool.acquire_index = 0;
    sem_pool.size = initial_size;
    sem_pool.semaphores = calloc(initial_size, sizeof(rt_vk_semaphore));
    if (!sem_pool.semaphores)
        return (rt_create_vk_semaphore_pool_result){.result = RT_OUT_OF_MEMORY};

    for (uint32_t i = 0; i < initial_size; ++i) {
        VkSemaphoreTypeCreateInfo type_info = {
            .sType = VK_STRUCTURE_TYPE_SEMAPHORE_TYPE_CREATE_INFO,
            .semaphoreType = VK_SEMAPHORE_TYPE_TIMELINE,
            .initialValue = 0,
        };
        VkSemaphoreCreateInfo semaphore_info = {
            .sType = VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO,
            .flags = 0,
            .pNext = &type_info
        };

        VkResult res = vkCreateSemaphore(dev->device, &semaphore_info, dev->alloc_cb, &sem_pool.semaphores[i].semaphore);
        if (res != VK_SUCCESS) {
            for (uint32_t j = 0; j < i; ++j) {
                vkDestroySemaphore(dev->device, sem_pool.semaphores[i].semaphore, dev->alloc_cb);
            }
            free(sem_pool.semaphores);
            return (rt_create_vk_semaphore_pool_result){.result = rtVkResultToRTResult(res)};
        }
        sem_pool.semaphores[i].value = 0;
        sem_pool.semaphores[i].in_use = RT_FALSE;
    }

    return (rt_create_vk_semaphore_pool_result){
        .result = RT_SUCCESS,
        .semaphore_pool = sem_pool,
    };
}

void rtDestroyVkSemaphorePool(rt_vk_semaphore_pool *pool) {
    for (uint32_t i = 0; i < pool->size; ++i) {
        vkDestroySemaphore(pool->dev->device, pool->semaphores[i].semaphore, pool->dev->alloc_cb);
    }
    free(pool->semaphores);
}

rt_vk_semaphore *rtAcquireSemaphore(rt_vk_semaphore_pool *pool) {
    rtLockRead(&pool->resize_lock);
    uint32_t index = rtAtomic32Inc(&pool->acquire_index) % pool->size;
    if (rtAtomic32CAS(&pool->semaphores[index].in_use, RT_TRUE, RT_FALSE) == RT_FALSE) {
        /* Successfully acquired the semaphore */
        rtUnlockRead(&pool->resize_lock);
        return &pool->semaphores[index];
    }

    /* We need to resize the pool */
    uint32_t pre_resize_size = pool->size;
    rtUnlockRead(&pool->resize_lock);
    rtLockWrite(&pool->resize_lock);
    if (pool->size > pre_resize_size) {
        /* Someone else was faster. Just try again */
        rtUnlockWrite(&pool->resize_lock);
        return rtAcquireSemaphore(pool);
    }
    uint32_t new_size = pre_resize_size * 2;
    rt_vk_semaphore *tmp = realloc(pool->semaphores, sizeof(rt_vk_semaphore) * new_size);
    if (!tmp) {
        rtUnlockWrite(&pool->resize_lock);
        rtReportError("VK", "Tried to grow the semaphore pool, but ran out of memory.");
        return NULL;
    }
    pool->semaphores = tmp;
    /* Create new semaphores */
    for (uint32_t i = pre_resize_size; i < new_size; ++i) {
        VkSemaphoreTypeCreateInfo type_info = {
            .sType = VK_STRUCTURE_TYPE_SEMAPHORE_TYPE_CREATE_INFO,
            .semaphoreType = VK_SEMAPHORE_TYPE_TIMELINE,
            .initialValue = 0,
        };
        VkSemaphoreCreateInfo semaphore_info = {
            .sType = VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO,
            .flags = 0,
            .pNext = &type_info
        };

        VkResult res = vkCreateSemaphore(pool->dev->device, &semaphore_info, pool->dev->alloc_cb, &pool->semaphores[i].semaphore);
        if (res != VK_SUCCESS) {
            for (uint32_t j = pre_resize_size; j < i; ++j) {
                vkDestroySemaphore(pool->dev->device, pool->semaphores[i].semaphore, pool->dev->alloc_cb);
            }
            rtReportError("VK", "Tried to grow the semaphore pool, but failed to create new semaphores");
        }
        pool->semaphores[i].value = 0;
        pool->semaphores[i].in_use = RT_FALSE;
    }
    pool->size = new_size;
    rtUnlockWrite(&pool->resize_lock);
    return rtAcquireSemaphore(pool);
}

void rtReleaseSemaphore(rt_vk_semaphore *semaphore) {
    rtAtomic32Exchange(&semaphore->in_use, RT_FALSE);
}
