#ifndef RT_ATOMICS_H
#define RT_ATOMICS_H

/* Macros & helpers for atomic instructions */

#ifdef _MSC_VER

/* Increment and decrement return the new value */

#define rtAtomic32Inc(pa) _InterlockedIncrement((volatile long *)(pa))
#define rtAtomic64Inc(pa) _InterlockedIncrement64((volatile __int64 *)(pa))
#define rtAtomic32Dec(pa) _InterlockedDecrement((volatile long *)(pa))
#define rtAtomic64Dec(pa) _InterlockedDecrement64((volatile __int64 *)(pa))

#define rtAtomic32FetchAdd(pa, value) _InterlockedExchangeAdd((volatile long *)(pa), (LONG)(value))
#define rtAtomic64FetchAdd(pa, value)                                                              \
    _InterlockedExchangeAdd64((volatile __int64 *)(pa), (LONG)(value))

#if defined(_M_ARM) || defined(_M_ARM64)
#define rtAtomic32ExchangeAcq(_pDest, _NewVal)                                                     \
    _InterlockedExchange_acq((volatile long *)(_pDest), (_NewVal))
#define rtAtomic32ExchangeRel(_pDest, _NewVal)                                                     \
    _InterlockedExchange_rel((volatile long *)(_pDest), (_NewVal))
#define rtAtomic32CASAcq(_pDest, _NewVal, _Compare)                                                \
    _InterlockedCompareExchange_acq((volatile long *)(_pDest), (_NewVal), (_Compare))
#define rtAtomic32CASRel(_pDest, _NewVal, _Compare)                                                \
    _InterlockedCompareExchange_rel((volatile long *)(_pDest), (_NewVal), (_Compare))
#else
/* x64/86 does not have acquire/release versions of these */
#define rtAtomic32ExchangeAcq(_pDest, _NewVal)                                                     \
    _InterlockedExchange((volatile long *)(_pDest), (_NewVal))
#define rtAtomic32ExchangeRel(_pDest, _NewVal)                                                     \
    _InterlockedExchange((volatile long *)(_pDest), (_NewVal))
#define rtAtomic32CASAcq(_pDest, _NewVal, _Compare)                                                \
    _InterlockedCompareExchange((volatile long *)(_pDest), (_NewVal), (_Compare))
#define rtAtomic32CASRel(_pDest, _NewVal, _Compare)                                                \
    _InterlockedCompareExchange((volatile long *)(_pDest), (_NewVal), (_Compare))
#endif
#define rtAtomic32Exchange(_pDest, _NewVal)                                                        \
    _InterlockedExchange((volatile long *)(_pDest), (_NewVal))
#define rtAtomic32CAS(_pDest, _NewVal, _Compare)                                                   \
    _InterlockedCompareExchange((volatile long *)(_pDest), (_NewVal), (_Compare))

#elif defined(__GNUC__) || defined(__clang__)

#define rtAtomic32Inc(pa) __atomic_add_fetch((pa), 1, __ATOMIC_SEQ_CST)
#define rtAtomic64Inc(pa) __atomic_add_fetch((pa), 1LL, __ATOMIC_SEQ_CST)
#define rtAtomic32Dec(pa) __atomic_sub_fetch((pa), 1, __ATOMIC_SEQ_CST)
#define rtAtomic64Dec(pa) __atomic_sub_fetch((pa), 1LL, __ATOMIC_SEQ_CST)

#define rtAtomic32FetchAdd(pa, value) __atomic_fetch_add((pa), (value), __ATOMIC_SEQ_CST)
#define rtAtomic64FetchAdd(pa, value) __atomic_fetch_add((pa), (value), __ATOMIC_SEQ_CST)

/* CAS "implementations" to make it conform to our expected api, i.e. return the original value of
 * *dest */
#define __RT_POSIX_CAS_N_IMPL(_T, _Name)                                                           \
    static RT_INLINE _T _Name(volatile _T *dest, _T new_val, _T compare, int memorder) {           \
        _T original = compare;                                                                     \
        if (!__atomic_compare_exchange_n(dest,                                                     \
                                         &compare,                                                 \
                                         new_val,                                                  \
                                         false,                                                    \
                                         memorder,                                                 \
                                         __ATOMIC_ACQUIRE))                                        \
            original = compare; /* Overwritten on failure with the original value */               \
        return original;                                                                           \
    }

__RT_POSIX_CAS_N_IMPL(int, __rtPOSIXCASInt)
__RT_POSIX_CAS_N_IMPL(long, __rtPOSIXCASLong)
__RT_POSIX_CAS_N_IMPL(uint32_t, __rtPOSIXCASUint32)

#define rtAtomic32ExchangeAcq(_pDest, _NewVal)                                                     \
    __atomic_exchange_n((_pDest), (_NewVal), __ATOMIC_ACQUIRE)
#define rtAtomic32ExchangeRel(_pDest, _NewVal)                                                     \
    __atomic_exchange_n((_pDest), (_NewVal), __ATOMIC_RELEASE)
#define rtAtomic32CASAcq(_pDest, _NewVal, _Compare)                                                \
    _Generic((_NewVal), int: __rtPOSIXCASInt, long: __rtPOSIXCASLong, uint32_t: __rtPOSIXCASUint32)((_pDest), (_NewVal), (_Compare), __ATOMIC_ACQUIRE)
#define rtAtomic32CASRel(_pDest, _NewVal, _Compare)                                                \
    _Generic((_NewVal), int: __rtPOSIXCASInt, long: __rtPOSIXCASLong, uint32_t: __rtPOSIXCASUint32)((_pDest), (_NewVal), (_Compare), __ATOMIC_RELEASE)
#define rtAtomic32Exchange(_pDest, _NewVal)                                                        \
    __atomic_exchange_n((_pDest), (_NewVal), __ATOMIC_SEQ_CST)
#define rtAtomic32CAS(_pDest, _NewVal, _Compare)                                                   \
    _Generic((_NewVal), int: __rtPOSIXCASInt, long: __rtPOSIXCASLong, uint32_t: __rtPOSIXCASUint32)((_pDest), (_NewVal), (_Compare), __ATOMIC_SEQ_CST)

#endif

#endif
